/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "find.c" - Services to find a specified directory entry        *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"
#include "unicode.h"
#include "wildcard.h"

#include "rtos.h"          // RTX_Sleep_Time()    - ad/031023

//#define DEBUG

#define LFN_FETCH_SLOTS 21

/* Searches an open directory for files matching the file specification */
/* and the search flags.                                                */
/* On success, returns 0, fills the FindData structure.                 */
/* Returns a negative error code on failure.                            */
int find(tFatVolume *V, int Fd, char *FileSpec, DWORD Flags,
         tFatFind *FindData, int UseLfn)
{
  /* To fetch the LFN we use a circular buffer Buffer.                   */
  /* A long file name is up to 255 characters long, so 20 entries are    */
  /* needed, plus one additional entry for the short name alias. We scan */
  /* the directory and place the entry that we read into the Buffer at   */
  /* the offset BufferPos. When we find a valid short name, we backtrace */
  /* the Buffer with LfnSlot to extract the long file name, if present.  */
  static  tFatDirEntry  Buffer[LFN_FETCH_SLOTS];
  tFatLfnEntry *LfnSlot;
  int           BufferPos = 0;

  int           Res, k;
  DWORD         NumRead;
  tFatFile     *Dir;
  int           Order;
  BYTE          Checksum;
  char          ShortName[13]; /* The short name entry */
  int           NamePos = 0; /* Points to the position inside LongName */
  static  WORD  LongName[FD32_MAX_LFN_LENGTH]; /* Found long name in UTF-16 */
  static  WORD  FileSpecUtf16[FD32_MAX_LFN_LENGTH]; /* FileSpec in UTF-16 */
  BYTE          AllowableAttr = (BYTE) Flags;
  BYTE          RequiredAttr  = (Flags >> 8) & 0xFF;
  #ifdef DEBUG
  char          Found8[FD32_MAX_LFN_LENGTH];
  #endif

  Dir = validate_fd(V, Fd);
  if (Dir == NULL) return FD32_ERROR_INVALID_HANDLE;

  if (!(Dir->DirEntry.Attr & FD32_ATTR_DIRECTORY))
    return FAT_RET_NOT_A_DIRECTORY;

  /* Convert FileSpec to Unicode as we'll do long name matching in Unicode */
  utf8_to_utf16(FileSpec, FileSpecUtf16);

  /* If at the end of the dir's cluster chain, we have not found the file */
  Res = fat_read(V, Fd, (BYTE *) &Buffer[0], 32, &NumRead);
  if (Res) return Res;
  while (NumRead != 0)
  {
    /* If the directory name starts with FAT_DIR_ENFOFDIR */
    /* we don't need to scan further.                     */
    if (Buffer[BufferPos].Name[0] == FAT_DIR_ENDOFDIR)
      return FD32_ERROR_NO_MORE_FILES;

    /* If the entry is marked as deleded we skip to the next entry */
    if (Buffer[BufferPos].Name[0] != FAT_DIR_FREE)
    {
      /* If we find a file without the Long Name attribute, it is a valid */
      /* short name. We then try to extract a long name before it.        */
      if (Buffer[BufferPos].Attr != FD32_ATTR_LONGNAME)
      {
// <not used>
        #ifdef FATLFN
        if (UseLfn)
        {
          Checksum = lfn_checksum(&Buffer[BufferPos]);
          LfnSlot  = (tFatLfnEntry *) &Buffer[BufferPos];
          NamePos  = 0;
          Order    = 0;
          do
          {
            if (--LfnSlot < (tFatLfnEntry *) Buffer) LfnSlot += LFN_FETCH_SLOTS;
            if (LfnSlot->Attr != FD32_ATTR_LONGNAME) break;
            if (++Order != (LfnSlot->Order & 0x1F)) break;
            if (Checksum != LfnSlot->Checksum) break;
            /* Ok, the LFN slot is valid, attach it to the long name */
            for (k = 0; k < 5; k++) LongName[NamePos++] = LfnSlot->Name0_4[k];
            for (k = 0; k < 6; k++) LongName[NamePos++] = LfnSlot->Name5_10[k];
            for (k = 0; k < 2; k++) LongName[NamePos++] = LfnSlot->Name11_12[k];
            /* If it's the last slot, check if it's our file */
            if (LfnSlot->Order & 0x40)
            {
              if (LongName[NamePos - 1] != 0x0000) LongName[NamePos] = 0x0000;
              if (((AllowableAttr | Buffer[BufferPos].Attr) == AllowableAttr)
               && ((RequiredAttr & Buffer[BufferPos].Attr) == RequiredAttr)
               && (unicode_stricmp_with_wildcards(LongName,FileSpecUtf16) == 0))
              {
                if (FindData)
                {
                  FindData->SfnEntry    = Buffer[BufferPos];
                  FindData->EntryOffset = Dir->TargetPos - NumRead;
                  FindData->LfnEntries  = Order;
                  absolute_name_to_short_name(FindData->ShortName,
                                              Buffer[BufferPos].Name);
                  utf16_to_utf8(LongName, FindData->LongName);
                }
                return 0;
              }
            }
          }
          while (!(LfnSlot->Order & 0x40));
        } /* if (UseLfn) */
        #endif /* FATLFN */
// </not used>
        /* Check if this short name is what we were searching for. Note */
        /* that this is only possible if AbsName[0] is not zero, i.e.   */
        /* the FileSpec to find fitted in a valid 8.3 name.             */
        absolute_name_to_short_name(ShortName, Buffer[BufferPos].Name);
        if (((AllowableAttr | Buffer[BufferPos].Attr) == AllowableAttr)
         && ((RequiredAttr & Buffer[BufferPos].Attr) == RequiredAttr)
         && (stricmp_with_wildcards(ShortName, FileSpec) == 0))
        {
          if (FindData)
          {
            FindData->SfnEntry    = Buffer[BufferPos];
            FindData->EntryOffset = Dir->TargetPos - NumRead;
            FindData->LfnEntries  = 0;
            absolute_name_to_short_name(FindData->ShortName,
                                        Buffer[BufferPos].Name);
            strcpy(FindData->LongName, FindData->ShortName);
          }
          return 0;
        }
      } /* if (!(Buffer[BufferPos].DIR_Attr & FAT_ATTR_LONGNAME)) */
    } /* if (Buffer[BufferPos].DIR_Name[0] != FAT_DIR_FREE) */
    if (++BufferPos == LFN_FETCH_SLOTS) BufferPos = 0;
    Res = fat_read(V, Fd, (BYTE *) &Buffer[BufferPos], 32, &NumRead);
    if (!BufferPos)
      RTX_Sleep_Time(1);     // let RTOS reset watchdog as this loop may take long - ad/031023
    if (Res) return Res;
  } /* while (NumRead != 0) */
  return FD32_ERROR_NO_MORE_FILES;
}


int fat_lfn_findfirst(void *P, char *FileSpec, DWORD Flags,
                      fd32_fs_lfnfind_t *FindData)
{
  tFatFile   *Dir;
  tFatVolume *V = (tFatVolume *) P;
  tFatFind    D;
  int         Fd, Res;
  static  char  Path[FD32_MAX_LFN_PATH_LENGTH];
  static  char  NameSpec[FD32_MAX_LFN_LENGTH];

  split_path(FileSpec, Path, NameSpec);
  Fd = fat_open(V, Path, FD32_OPEN_ACCESS_READ | FAT_OPEN_DIRECTORY,
                FD32_ATTR_NONE, 0, NULL, FD32_LFN);
  if (Fd < 0) return Fd;
  Dir = &V->Files[Fd];

  strcpy(Dir->FindFileSpec, NameSpec);
  Dir->FindFlags = Flags;

  Res = find(V, Fd, Dir->FindFileSpec, Flags, &D, FD32_LFN);
  if (Res) return Res;
  /* FIX ME: Should also support the Windows 64-bit time format (number of
   * 100ns intervals since 1/1/1601)
   */
  FindData->Attr   = D.SfnEntry.Attr;
  FindData->CTime  = D.SfnEntry.CrtTime + (((unsigned long)D.SfnEntry.CrtDate) << 16);
  FindData->ATime  = ((unsigned long)D.SfnEntry.LstAccDate) << 16;
  FindData->MTime  = D.SfnEntry.WrtTime + (((unsigned long) D.SfnEntry.WrtDate) << 16);
  FindData->SizeHi = 0;
  FindData->SizeLo = D.SfnEntry.FileSize;
  FindData->Vol = V;
  FindData->Handle = Fd;
  FindData->Flags = Flags;
  strcpy(FindData->LongName, D.LongName);
  strcpy(FindData->ShortName, D.ShortName);
  return 0;
}


int fat_lfn_findnext(fd32_fs_lfnfind_t *FindData)
{
  tFatFile   *Dir;
  tFatVolume *V = (tFatVolume *) FindData->Vol;
  tFatFind    D;
  int         Res;
  DWORD       Flags = FindData->Flags;
  int         Fd = FindData->Handle;

  if ((Dir = validate_fd(V, Fd)) == NULL) return FD32_ERROR_INVALID_HANDLE;
  /* FIX ME: Should mix previous flags and current flags */
  Res = find(V, Fd, Dir->FindFileSpec, Flags, &D, FD32_LFN);
  if (Res) return Res;
  /* FIX ME: Should also support the Windows 64-bit time format (number of
   * 100ns intervals since 1/1/1601)
   */
  FindData->Attr   = D.SfnEntry.Attr;
  FindData->CTime  = D.SfnEntry.CrtTime + (((unsigned long)D.SfnEntry.CrtDate) << 16);
  FindData->ATime  = ((unsigned long)D.SfnEntry.LstAccDate) << 16;
  FindData->MTime  = D.SfnEntry.WrtTime + (((unsigned long)D.SfnEntry.WrtDate) << 16);
  FindData->SizeHi = 0;
  FindData->SizeLo = D.SfnEntry.FileSize;
  strcpy(FindData->LongName, D.LongName);
  strcpy(FindData->ShortName, D.ShortName);
  return 0;
}


int fat_lfn_findclose(fd32_fs_lfnfind_t *FindData)
{
  tFatVolume *V = (tFatVolume *) FindData->Vol;
  int         Fd = FindData->Handle;
  tFatFile   *Dir = validate_fd(V, Fd);

  if (Dir == NULL) return FD32_ERROR_INVALID_HANDLE;
  if (!(Dir->DirEntry.Attr & FD32_ATTR_DIRECTORY)) return FAT_RET_NOT_A_DIRECTORY;
  return fat_close(V, Fd);
}


int fat_dos_findfirst(char VolumeNo, char *FileSpec, BYTE Attributes,
                      fd32_fs_dosfind_t *Dta)
{
  int         Res, Fd;
  tFatFile   *Dir;
  tFatVolume *V  = devices[VolumeNo];
  tDosFind   *Df = (tDosFind *) Dta;
  tFatFind    D;
  char        Path[FD32_MAX_SFN_PATH_LENGTH];
  char        Filename[FD32_MAX_SFN_PATH_LENGTH];

  split_path(FileSpec, Path, Filename);
  short_name_to_absolute_name(Filename, Df->SearchName);
  /* The Ralf Brown's list says that:
   * "for search attributes other than 08h, all files with at MOST the
   * specified combination of hidden, system, and directory attributes
   * will be returned.  Under DOS 2.x, searching for attribute 08h
   * (volume label) will also return normal files, while under DOS 3.0+
   * only the volume label (if any) will be returned."
   */
/*  if (Attributes == FD32_ATTR_VOLUMEID)
    Df->SearchFlags = FD32_FIND_ALLOW_VOLUMEID | FD32_FIND_REQUIRE_VOLUMEID;
  else
    Df->SearchFlags = Attributes | FD32_FIND_REQUIRE_NONE;*/

  // [KG] changed SearchFlags to Byte-Value, so attr_volumeid ist threatened like
  // any other attribute:
  Df->SearchFlags = Attributes;
  Df->Drive = VolumeNo;

  Fd = fat_open(V, Path, FD32_OPEN_ACCESS_READ | FAT_OPEN_DIRECTORY,
                FD32_ATTR_NONE, 0, NULL, FD32_NOLFN);
  if (Fd < 0) return Fd;
  Dir = &V->Files[Fd];
  Df->SearchDirStart = Dir->FirstCluster;
//  Df->SearchVol      = V;

  Res = find(V, Fd, Filename, Df->SearchFlags, &D, FD32_NOLFN);
  if (Res) { fat_close(V, Fd); return Res; }
  /* Update the finddata record */
  Df->SearchDirPos = Dir->TargetPos;
  Dta->Attr  = D.SfnEntry.Attr;
  Dta->MTime = D.SfnEntry.WrtTime;
  Dta->MDate = D.SfnEntry.WrtDate;
  Dta->Size  = D.SfnEntry.FileSize;
  strcpy(Dta->Name, D.ShortName);
  return fat_close(V, Fd);
}


int fat_dos_findnext(fd32_fs_dosfind_t *Dta)
{
  int       Res, Fd;
  tFatFile *Dir;
  tDosFind *Df = (tDosFind *) Dta;
  tFatFind  D;
  char      SearchName[15];
  absolute_name_to_short_name(SearchName, Df->SearchName);

  Fd = reopen_dir(devices[Df->Drive], FD32_OPEN_ACCESS_READ | FAT_OPEN_DIRECTORY,
                  Df->SearchDirStart);
  if (Fd < 0) return Fd;
  Dir = &devices[Df->Drive]->Files[Fd];
  Dir->TargetPos = Df->SearchDirPos;

  Res = find(devices[Df->Drive], Fd, SearchName, Df->SearchFlags, &D, FD32_NOLFN);
  if (Res) { fat_close(devices[Df->Drive], Fd); return Res; }
  /* Update the finddata record */
  Df->SearchDirPos = Dir->TargetPos;
  Dta->Attr  = D.SfnEntry.Attr;
  Dta->MTime = D.SfnEntry.WrtTime;
  Dta->MDate = D.SfnEntry.WrtDate;
  Dta->Size  = D.SfnEntry.FileSize;
  strcpy(Dta->Name, D.ShortName);
  return fat_close(devices[Df->Drive], Fd);
}

