/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "open.c" - Open, create and close a file (or even a directory  *
 *                    as a file) in any directory, allocating and freeing *
 *                    handles                                             *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"

//#define DEBUG


/* Searches for an unused file descriptor of the volume V.       */
/* Returns a valid file descriptor on success, or -1 on failure. */
/* Called by fat_open, reopen_dir and fat_fflush.                */
static int allocate_fd(tFatVolume *V)
{
  int k;
  for (k = 0; k < FAT_MAX_FILES; k++)
    if (!(V->Files[k].Mode & FAT_FMODE_OPENED))
    {
      V->Files[k].Mode = FAT_FMODE_OPENED;
      return k;
    }
  return -1; /* File descriptor not available */
}


/* Marks as unused an open file descriptor of the volume V.   */
/* Returns 0 on success, or a negative error code on failure. */
/* Called by fat_open, reopen_dir, fat_fflush and fat_close.  */
static int free_fd(tFatVolume *V, int Fd)
{
  tFatFile *F = validate_fd(V, Fd);
  if (F == NULL) return FD32_ERROR_INVALID_HANDLE;
  F->Mode = 0;
  return 0;
}


/* Returns the file descriptor of the first time the file has been */
/* opened, or -1 if the file is not yet open.                      */
/* Called by set_opening_mode.                                     */
static int first_open_fd(tFatVolume *V, DWORD FirstCluster)
{
  int k;
  for (k = 0; k < FAT_MAX_FILES; k++)

    if ((V->Files[k].FirstCluster == FirstCluster)
     && (V->Files[k].Mode & FAT_FMODE_FIRST_OPEN)) return k;
  return -1; /* File is currently not open */
}


/* Splits a full valid path name (path + file name) into its path  */
/* and file name components.                                       */
/* Called by fat_open, fat_unlink (unlink.c), fat_lfn_findfirst    */
/* (find.c), fat_dos_findfirst (find.c) and fat_rename (rename.c). */
void split_path(char *FullPath, char *Path, char *Name)
{
  char *NameStart;
  char *s;
  
  /* Move to the end of the FullPath string */
  for (s = FullPath; *s != 0; s++);
  /* Now move backward until '\' is found. The name starts next the '\' */
  for (; (*s != '\\') && (s != FullPath); s--);
  if (*s == '\\') s++;
  /* And copy the string from this point until the end into Name */
  for (NameStart = s; *s != 0; s++, Name++) *Name = *s;
  *Name = 0;
  /* Finally we copy from the beginning to the filename into Path, with
   * the trailing slash.
   */
  for (s = FullPath; s < NameStart; s++, Path++) *Path = *s;
  *Path = 0;
}


/* Creates a zero length file, allocating a new directory entry, and */
/* overwriting the parent directory file structure with its status.  */
/* Returns 0 on success, or a negative error code on failure.        */
/* Called by open_using_the_same_fd.                                 */
static int fat_creat(tFatVolume *V, int Fd, char *Path, char *Name,
                     BYTE Attributes, WORD AliasHint, int UseLfn)
{
  tFatFile     *F = &V->Files[Fd];
  int           Res;
  tFatDirEntry  D;
  DWORD         DirEntryOffset;
  int           DirLfnEntries;

  /* Initialize the directory entry */
  D.Attr         = Attributes;
  D.NTRes        = 0;
  update_date_and_time(&D.CrtTime, &D.CrtDate, &D.CrtTimeTenth);
  D.LstAccDate   = D.CrtDate;
  D.FstClusHI    = 0;
  D.WrtTime      = D.CrtTime;
  D.WrtDate      = D.CrtDate;
  D.FstClusLO    = 0;
  D.FileSize     = 0;

  F->TargetPos = 0;
  /* Switch the access mode to READWRITE for the following operation */
  F->Mode      = (F->Mode & ~FD32_OPEN_ACCESS_MASK) | FD32_OPEN_ACCESS_READWRITE;
  #ifdef FATLFN
  if (UseLfn)
    Res = allocate_lfn_dir_entries(V, Fd, &D, Path, Name, &DirEntryOffset,
                                   &DirLfnEntries, AliasHint);
   else
    Res = allocate_sfn_dir_entry(V, Fd, &D, Name, &DirEntryOffset);
  #else
  Res = allocate_sfn_dir_entry(V, Fd, &D, Name, &DirEntryOffset);
  #endif
  if (Res) return Res;
  /* Now replace the content of the file structure of the directory */
  /* containing the new file with the status of the new file.       */
  F->Mode            = FAT_FMODE_OPENED;
  F->FirstCluster    = 0;
  F->ParentDirEntry  = F->DirEntry; /* Save the parent dir entry...         */
  F->DirEntry        = D;           /* ...and replace it with the new entry */
  F->DirEntryOffset  = DirEntryOffset;
  F->DirLfnEntries   = DirLfnEntries;
  /* Seek to the beginning */
  F->TargetPos       = 0;
  F->FilePos         = 0;
  F->Cluster         = F->FirstCluster;
  F->PrevCluster     = 0;
  F->SectorInCluster = 0;
  F->ByteInSector    = 0;

  /* The opening mode */
  F->Mode &= ~FD32_OPEN_ACCESS_MASK;
  return 0;
}


/* Sets the file opening mode according to the fat_open() parameter.      */
/* All file structure's flags except FAT_FMODE_OPENED should be clear.    */
/* Returns 0 on success, or a negative error code on failure.             */
/* Called by open_using_the_same_fd, fat_open, reopen_dir and fat_fflush. */
static int set_opening_mode(tFatVolume *V, int Fd, DWORD Mode)
{
  tFatFile *F = &V->Files[Fd];
  int       Res;

  /* Directory consistency check */
  if (F->DirEntry.Attr & FD32_ATTR_DIRECTORY)
  {
    if (!(Mode & FAT_OPEN_DIRECTORY)) return FD32_ERROR_ACCESS_DENIED;
  }
  else
  {
    if (Mode & FAT_OPEN_DIRECTORY) return FAT_RET_NOT_A_DIRECTORY;
  }

  F->Mode |= Mode;

  /* Check if we can open for writing */
  switch (Mode & FD32_OPEN_ACCESS_MASK)
  {
    case FD32_OPEN_ACCESS_WRITE     :
    case FD32_OPEN_ACCESS_READWRITE :
      if (F->DirEntry.Attr & FD32_ATTR_READONLY) {
        return FD32_ERROR_ACCESS_DENIED;
      }
  }

  /* Sharing behaviour as in DOS 7.x */
  Res = first_open_fd(V, F->FirstCluster);
  if (Res != -1)
  {
    Res = check_for_sharing_conflict(V, Fd, Res);
    if (Res) return Res;
  }
  else F->Mode |= FAT_FMODE_FIRST_OPEN;

  return 0;
}


/* Opens a file overwriting the file structure used for the parent */
/* directory. Used to descend the path.                            */
/* On success, returns 0 and fills Result with the action taken.   */
/* Returns a negative error code on failure.                       */
/* Called by fat_open                                              */
static int open_using_the_same_fd(tFatVolume *V, int Fd, char *Path,
                                  char *Name, DWORD Mode, BYTE Attributes,
                                  WORD AliasHint, int *Result, int UseLfn)
{
  tFatFile *F = &V->Files[Fd];
  static  tFatFind  D;
  int       Res;

  /* Search for the file FileName in the directory DirHandle */
  Res = find(V, Fd, Name, FD32_FIND_REQUIRE_NONE | FD32_FIND_ALLOW_ALL,
             &D, UseLfn);
  /* Report immediatly any error different from "no match found" */
  if (Res && (Res != FD32_ERROR_NO_MORE_FILES)) return Res;

  /* We may need to create the file if it does not exist */
  if (Res == FD32_ERROR_NO_MORE_FILES)
  {
    if (!(Mode & FD32_OPEN_NOTEXIST_CREATE))
    {
      if (Mode & FAT_OPEN_DIRECTORY) return FD32_ERROR_PATH_NOT_FOUND;
                                else return FD32_ERROR_FILE_NOT_FOUND;
    }
    Res = fat_creat(V, Fd, Path, Name, Attributes, AliasHint, UseLfn);
    if (Res) return Res;

    if (Result) *Result = FD32_OPEN_RESULT_CREATED;
    return set_opening_mode(V, Fd, Mode);
  }

  /* If we arrive here, the file already exists */

  /* FIX ME: The file system specification says that if both           */
  /*         FD32_OPEN_EXIST_OPEN and FD32_OPEN_EXIST_REPLACE are set, */
  /*         FD32_OPEN_EXIST_OPEN shall be ignored.                    */
  /*         By now we actually do the opposite...                     */
  /*         However, this behaviour should be compared with Windows.  */
  if (!(Mode & FD32_OPEN_EXIST_OPEN)) return FD32_ERROR_ACCESS_DENIED;

  /* Now replace the content of the file structure of the directory */
  /* containing the file we found with the status of the file.      */
  F->Mode           = FAT_FMODE_OPENED;
  F->FirstCluster   = (((unsigned long) D.SfnEntry.FstClusHI) << 16) + D.SfnEntry.FstClusLO;
  F->ParentDirEntry = F->DirEntry;
  F->DirEntry       = D.SfnEntry;
  F->DirEntryOffset = D.EntryOffset;
  F->DirLfnEntries  = D.LfnEntries;
  /* Check if we are going to open the root directory, identified as a dir */
  /* starting at cluster 0, like in "..", even if the volume is FAT32.     */
  if ((F->FirstCluster == 0) && (F->DirEntry.Attr & FD32_ATTR_DIRECTORY))
  {
    if (V->FatType != FATTYPE_FAT32) F->Mode |= FAT_FMODE_ROOT;
                                else F->FirstCluster = V->Bpb.BPB_RootClus;
  }
  /* Seek to the beginning */
  F->TargetPos       = 0;
  F->FilePos         = 0;
  F->Cluster         = F->FirstCluster;
  F->PrevCluster     = 0;
  F->SectorInCluster = 0;
  F->ByteInSector    = 0;

  Res = set_opening_mode(V, Fd, Mode);
  if (Res) return Res;

  /* Truncate the file to zero length if required */
  if (Result) *Result = FD32_OPEN_RESULT_OPENED;
  if (Mode & FD32_OPEN_EXIST_REPLACE)
  {
    Res = fat_write(V, Fd, NULL, 0, NULL);
    if (Res) return Res;
    if (Result) *Result = FD32_OPEN_RESULT_REPLACED;
  }
  return 0;
}


/* Opens, creates or replaces a file.                            */
/* On success, returns 0 and fills Result with the action taken. */
/* Returns a negative error code on failure.                     */
/* This is a public driver function.                             */
int fat_open(void *P, char *FileName, DWORD Mode, WORD Attr,
             WORD AliasHint, int *Result, int UseLfn)
{
  int         Fd;
  int         Res;
  static  char  PartialName[FD32_MAX_LFN_LENGTH];
  static  char  Path[FD32_MAX_LFN_PATH_LENGTH];
  static  char  Name[FD32_MAX_LFN_LENGTH];
  char       *pPath = Path;
  char       *pPartialName;
  tFatFile   *F;
  tFatVolume *Vol = (tFatVolume *) P;

  split_path(FileName, Path, Name);
  #ifdef DEBUG
  printf("OPEN in device \"%s\"\n", Vol->DevName);
  printf("Passed file name is \"%s\"\n", FileName);
  printf("Opening file \"%s\" in directory \"%s\"\n", Name, Path);
  #endif

  /* Allocate a file descriptor for the file.  */
  /* It will be also used to descend the path. */
  Fd = allocate_fd(Vol);
  if (Fd == -1) return FD32_ERROR_TOO_MANY_OPEN_FILES;
  #ifdef DEBUG
  printf(" Fd %i allocated.\n", Fd);
  #endif
  F = &Vol->Files[Fd];
  #ifdef DEBUG
  printf(" memsetting...\n");
  #endif
  memset(&F->DirEntry, 0, sizeof(F->DirEntry));
  #ifdef DEBUG
  printf(" Starting to search.\n");
  #endif
  F->DirEntry.Attr = FD32_ATTR_DIRECTORY;
  /* Start the search from the root dir or from the current dir */
  if (*pPath == '\\')
  {
    pPath++;
    if (Vol->FatType != FATTYPE_FAT32) F->Mode |= FAT_FMODE_ROOT;
    F->FirstCluster = Vol->Bpb.BPB_RootClus; /* Zero for FAT12/FAT16 */
  }
  else
  {
    F->FirstCluster = Vol->CurDirCluster;
    if ((Vol->FatType != FATTYPE_FAT32) && (F->FirstCluster == 0))
      F->Mode |= FAT_FMODE_ROOT;
  }

  // [KG] FIX: directory entries of files created in root-directory have been created
  //      but not updated, after write access.
  F->DirEntry.FstClusHI = F->FirstCluster >> 16;
  F->DirEntry.FstClusLO = F->FirstCluster & 0xffff;
  // /FIX

  F->TargetPos       = 0;
  F->FilePos         = 0;
  F->Cluster         = F->FirstCluster;
  F->PrevCluster     = 0;
  F->SectorInCluster = 0;
  F->ByteInSector    = 0;

  Res = set_opening_mode(Vol, Fd, FD32_OPEN_ACCESS_READ | FAT_OPEN_DIRECTORY);
  if (Res) { free_fd(Vol, Fd); return Res; }

  #ifdef DEBUG
  printf(" Descending the path...\n");
  #endif
  /* Descend the path */
  while (*pPath)
  {
    pPartialName = PartialName;
    while (*pPath != '\\') *(pPartialName++) = *(pPath++);
    pPath++;
    *pPartialName = 0;

    #ifdef DEBUG
    printf(" Partial : #%s#\n", PartialName);
    #endif

    /* Here the path will never be used because fat_creat is not allowed */
    Res = open_using_the_same_fd(Vol, Fd, NULL, PartialName,
                                 FD32_OPEN_EXIST_OPEN |
                                 FD32_OPEN_ACCESS_READ | FAT_OPEN_DIRECTORY,
                                 FD32_ATTR_NONE, 0, Result, UseLfn);
    if (Res) { free_fd(Vol, Fd); return Res; }
  }
  #ifdef DEBUG
  printf(" Opening the actual file...\n");
  #endif
  /* And finally open the file */
  if (*Name)
  {
    if (!(Mode & FD32_OPEN_USE_ALIAS_HINT)) AliasHint = 1;
    Res = open_using_the_same_fd(Vol, Fd, Path, Name, Mode, (BYTE) Attr,
                                 AliasHint, Result, UseLfn);
    if (Res) { free_fd(Vol, Fd); return Res; }
  }
  else
  {
    /* If no file name has been provided, apply the mode to the directory */
    Res = set_opening_mode(Vol, Fd, Mode);
    if (Res) { free_fd(Vol, Fd); return Res; }
  }
  #ifdef DEBUG
  printf(" File successfully open on Fd %i\n", Fd);
  #endif
  return Fd;
}


/* Opens a directory knowing its first cluster.                    */
/* Returns a file descriptor on success, or a negative error code. */
/* Called by fat_dos_findnext to reopen the search directory.      */
int reopen_dir(tFatVolume *V, DWORD Mode, DWORD FirstCluster)
{
  tFatFile *F;
  SDWORD    Fd, Res;
  
  Fd = allocate_fd(V);
  if (Fd == -1) return FD32_ERROR_TOO_MANY_OPEN_FILES;
  F = &V->Files[Fd];

  if (FirstCluster == 0) F->Mode |= FAT_FMODE_ROOT;
  F->FirstCluster    = FirstCluster;
  F->TargetPos       = 0;
  F->FilePos         = 0;
  F->Cluster         = F->FirstCluster;
  F->PrevCluster     = 0;
  F->SectorInCluster = 0;
  F->ByteInSector    = 0;
  
  Res = set_opening_mode(V, Fd, Mode);
  if (Res) { free_fd(V, Fd); return Res; }
  return Fd;
}


/* Flushes the content of volume's buffers and updates the directory entry. */
/* Returns 0 on success, or a negative error code on failure.               */
/* This is a public driver function.                                        */
int fat_fflush(void *P, int Fd)
{
  int         Res;
  int         DirFd;
  tFatFile   *F;
  tFatFile   *Dir;
  tFatVolume *V = (tFatVolume *) P;

  F = validate_fd(V, Fd);
  if (F == NULL) return FD32_ERROR_INVALID_HANDLE;

  /* Update parent directory entry only if the file itself is not the root */
  if (!(F->Mode & FAT_FMODE_ROOT) && (F->FirstCluster != V->Bpb.BPB_RootClus))
  {
    DirFd = allocate_fd(V);
    if (DirFd == -1) return FD32_ERROR_TOO_MANY_OPEN_FILES;
    Dir = &V->Files[DirFd];

    Dir->DirEntry        = F->ParentDirEntry;
    Dir->FirstCluster    = (((unsigned long) Dir->DirEntry.FstClusHI) << 16)
                         +  Dir->DirEntry.FstClusLO;
    if (Dir->FirstCluster == 0) Dir->Mode |= FAT_FMODE_ROOT;
    Dir->TargetPos       = F->DirEntryOffset;
    Dir->FilePos         = 0;
    Dir->Cluster         = Dir->FirstCluster;
    Dir->PrevCluster     = 0;
    Dir->SectorInCluster = 0;
    Dir->ByteInSector    = 0;

    Res = set_opening_mode(V, DirFd,
                           FD32_OPEN_ACCESS_WRITE | FAT_OPEN_DIRECTORY);
    if (Res) { free_fd(V, DirFd); return Res; }

    /* Write the file's directory entry into the file's directory */
    Res = fat_write(V, DirFd, (BYTE *) &F->DirEntry, 32, NULL);
    if (Res)
    {
      int Res1 = flush_all_buffers(V);
      if (Res1) return Res1;
      free_fd(V, DirFd);
      return Res;
    }
    free_fd(V, DirFd);
  }
  Res = flush_all_buffers(V);
  return Res;
}


/* Flushes buffers, updates directory entry, releases the file structure. */
/* Returns 0 on success, or a negative error code on failure.             */
/* This is a public driver function.                                      */
int fat_close(void *P, int Fd)
{
  int Res;
  tFatVolume *V = (tFatVolume *) P;
  if (validate_fd(V, Fd) == NULL) return FD32_ERROR_INVALID_HANDLE;
  Res = fat_fflush(V, Fd);
  if (Res) return Res;
  free_fd(V, Fd);
  return 0;
}
