/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "seek.c" - Move the actual file position to the target         *
 *                    position specified by the lseek primitive           *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"

//#define DEBUG


/* Moves the file pointer (the target position for read and write  */
/* operations) for a file, according to the origin:                */
/*  0 - the file pointer is moved to the offset specified          */
/*  1 - the file pointer is moved relative to its current position */
/*  2 - the file pointer is moved from the end of the file (note   */
/*      that this Origin is not valid for directories, since the   */
/*      FileSize for a directory is always zero).                  */
/* On success, returns 0 and, if Result is not NULL, fills Result  */
/* with the new absolute target position.                          */
/* This is a public driver function.                               */
int fat_lseek(void *P, int Fd, SQWORD Offset, int Origin, SQWORD *Result)
{
  tFatVolume *V = (tFatVolume *) P;
  tFatFile   *F = validate_fd(V, Fd);
  if (F == NULL) return FD32_ERROR_INVALID_HANDLE;
  switch (Origin)
  {
    case FD32_SEEK_SET : F->TargetPos = Offset; break;
    case FD32_SEEK_CUR : F->TargetPos += Offset; break;
    case FD32_SEEK_END : if (F->DirEntry.Attr & FD32_ATTR_DIRECTORY)
                           return FD32_ERROR_INVALID_FUNCTION;
                         F->TargetPos = F->DirEntry.FileSize + Offset;
                         break;
    default            : return FD32_ERROR_INVALID_FUNCTION;
  }
  if (Result) *Result = F->TargetPos;
  return 0;
}


/* The number of the first free data cluster found in the FAT volume,  */
/* starting from the cluster Vol->FSI_Nxt_Free, if this value is       */
/* available. If Vol->FSI_Nxt_Free is invalid, the FAT is scanned and  */
/* Vol->FSI_Nxt_Free is set to the first free cluster found.           */
/* On success returns 0 and fills Cluster with the first free cluster  */
/* number. Returns a negative error code on failure.                   */
/* Called by allocate_and_link_new_cluster and allocate_first_cluster. */
static int first_free_cluster(tFatVolume *V, DWORD *Cluster)
{
  DWORD k, Value;
  int   Res;

  /* Check if we know where the next free cluster is and if that value is */
  /* in the range of valid data clusters. If not we start searching from  */
  /* cluster 2, the first data cluster.                                   */
  if ((V->FSI_Nxt_Free == 0xFFFFFFFF)
   || (V->FSI_Nxt_Free < 2)
   || (V->FSI_Nxt_Free > V->DataClusters + 1)) V->FSI_Nxt_Free = 2;

  switch (V->FatType)
  {
    case FATTYPE_FAT12 : for (k = V->FSI_Nxt_Free; k < V->DataClusters + 2; k++)
                         {
                           Res = fat12_read_entry(V, k, 0, &Value);
                           if (Res) return Res;
                           if (Value == 0)
                           {
                             *Cluster = V->FSI_Nxt_Free = k;
                             return 0;
                           }
                         }
                         V->FSI_Nxt_Free = 0xFFFFFFFF;
                         return FD32_ERROR_DISK_FULL;
    case FATTYPE_FAT16 : for (k = V->FSI_Nxt_Free; k < V->DataClusters + 2; k++)
                         {
                           Res = fat16_read_entry(V, k, 0, &Value);
                           if (Res) return Res;
                           if (Value == 0)
                           {
                             *Cluster = V->FSI_Nxt_Free = k;
                             return 0;
                           }
                         }
                         V->FSI_Nxt_Free = 0xFFFFFFFF;
                         return FD32_ERROR_DISK_FULL;
    case FATTYPE_FAT32 : for (k = V->FSI_Nxt_Free; k < V->DataClusters + 2; k++)
                         {
                           Res = fat32_read_entry(V, k, 0, &Value);
                           if (Res) return Res;
                           if (Value == 0)
                           {
                             *Cluster = V->FSI_Nxt_Free = k;
                             return 0;
                           }
                         }
                         V->FSI_Nxt_Free = 0xFFFFFFFF;
                         return FD32_ERROR_DISK_FULL;
    default            : /* should never arrive here */
                         return FAT_RET_CRITICAL_ERROR;
  }
}


/* Allocates a free cluster and links it to the PrevCluster of a file,   */
/* that is expected to be set to the last allocated cluster (that marked */
/* with EOC). Then mark that new cluster entry with EOC.                 */
/* On success returns 0 and fills Cluster with the new cluster number.   */
/* Returns a negative error code on failure.                             */
/* Called by move_to_targetpos.                                          */
static int allocate_and_link_new_cluster(tFatVolume *V, int Fd, DWORD *Cluster)
{
  int   k, Res;
  DWORD NewCluster;

  Res = first_free_cluster(V, &NewCluster);
  if (Res) return Res;

  /* Update every FAT in the volume */
  for (k = 0; k < V->Bpb.BPB_NumFATs; k++)
  {
    /* Link the current cluster to the new cluster allocated */
    switch (V->FatType)
    {
      case FATTYPE_FAT12 : Res = fat12_write_entry(V, V->Files[Fd].PrevCluster, k, NewCluster); break;
      case FATTYPE_FAT16 : Res = fat16_write_entry(V, V->Files[Fd].PrevCluster, k, NewCluster); break;
      case FATTYPE_FAT32 : Res = fat32_write_entry(V, V->Files[Fd].PrevCluster, k, NewCluster); break;
      default            : /* should never arrive here */
                           return FAT_RET_CRITICAL_ERROR;
    }
    if (Res) return Res;

    /* Mark the new cluster allocated with EOC */
    switch (V->FatType)
    {
      case FATTYPE_FAT12 : Res = fat12_write_entry(V, NewCluster, k, 0x0FFF); break;
      case FATTYPE_FAT16 : Res = fat16_write_entry(V, NewCluster, k, 0xFFFF); break;
      case FATTYPE_FAT32 : Res = fat32_write_entry(V, NewCluster, k, 0x0FFFFFFF); break;
      default            : /* should never arrive here */
                           return FAT_RET_CRITICAL_ERROR;
    }
    if (Res) return Res;
  }
  *Cluster = NewCluster;
  return 0;
}


/* Allocates a free cluster to be used as first cluster of a file, that    */
/* is supposed to be a zero length file. Then marks that new cluster entry */
/* with EOC.                                                               */
/* On success returns 0 and fills Cluster with the new cluster number.     */
/* Returns a negative error code on failure.                               */
/* Called by move_to_targetpos.                                            */
static int allocate_first_cluster(tFatVolume *V, int Fd, DWORD *Cluster)
{
  int   k, Res;
  DWORD NewCluster;

  Res = first_free_cluster(V, &NewCluster);
  if (Res) return Res;

  /* Make the new cluster allocated the first cluster */
  V->Files[Fd].FirstCluster = NewCluster;
  /* Update the directory entry content */
  V->Files[Fd].DirEntry.FstClusHI = NewCluster >> 16;
  V->Files[Fd].DirEntry.FstClusLO = (WORD) NewCluster;

  /* Update every FAT in the volume */
  for (k = 0; k < V->Bpb.BPB_NumFATs; k++)
  {
    /* Mark the new cluster allocated with EOC */
    switch (V->FatType)
    {
      case FATTYPE_FAT12 : Res = fat12_write_entry(V, NewCluster, k, 0x0FFF); break;
      case FATTYPE_FAT16 : Res = fat16_write_entry(V, NewCluster, k, 0xFFFF); break;
      case FATTYPE_FAT32 : Res = fat32_write_entry(V, NewCluster, k, 0x0FFFFFFF); break;
      default            : /* should never arrive here */
                           return FAT_RET_CRITICAL_ERROR;
    }
    if (Res) return Res;
  }
  *Cluster = NewCluster;
  return 0;
}


/* Returns nonzero if the pointer of the file has reached the End Of File. */
/* Called by move_to_targetpos.                                            */
static int end_of_file(tFatVolume *V, int Fd)
{
  tFatFile *F = &V->Files[Fd];

  if (F->Mode & FAT_FMODE_ROOT)
  {
    /* Case 1: the end of a FAT12/FAT16 root directory */
    if (F->FilePos >= V->Bpb.BPB_RootEntCnt * 32) return 1;
  }
  else
  {
    /* Case 2: the file is not a directory, and we are at the last byte */
    if ((F->FilePos >= F->DirEntry.FileSize)
     && (!(F->DirEntry.Attr & FD32_ATTR_DIRECTORY))) return 1;

    /* Case 3: we arrived at the last cluster, that marked with EOC */
    switch (V->FatType)
    {
      case FATTYPE_FAT12 : if (FAT12_EOC(F->Cluster)) return 1;
      case FATTYPE_FAT16 : if (FAT16_EOC(F->Cluster)) return 1;
      case FATTYPE_FAT32 : if (FAT32_EOC(F->Cluster)) return 1;
    }
  }
  return 0;
}

/* Move the file pointer forward by one byte, jumping to next sector */
/* or to next cluster if required.                                   */
/* Returns 0 on success, or a negative error code on failure.        */
/* Called by move_to_targetpos.                                      */

//[KG] SPEEDUP: Try to move by MORE THAN ONE BYTE... (IncBy = Count of Bytes to move)
static int advance_byte_position(tFatVolume *V, int Fd, unsigned int IncBy)
{
  int       Res;
  tFatFile *F = &V->Files[Fd];

  F->FilePos+=IncBy;
  F->ByteInSector+=IncBy;
  if (F->ByteInSector == V->Bpb.BPB_BytsPerSec)
  {
//printf("\nJump to sector: %lu\n", F->SectorInCluster);
    F->ByteInSector = 0;
    F->SectorInCluster++;

    /* If the file is not a FAT12/FAT16 root, if we are at the cluster */
    /* boundary we jump to the next cluster of the file chain.         */
    if ((!(F->Mode & FAT_FMODE_ROOT)) &&
        (F->SectorInCluster == V->Bpb.BPB_SecPerClus))
    {
      F->SectorInCluster = 0;
      F->PrevCluster     = F->Cluster;
      switch (V->FatType)
      {
        case FATTYPE_FAT12 : Res = fat12_read_entry(V, F->Cluster, 0, &F->Cluster); break;
        case FATTYPE_FAT16 : Res = fat16_read_entry(V, F->Cluster, 0, &F->Cluster); break;
        case FATTYPE_FAT32 : Res = fat32_read_entry(V, F->Cluster, 0, &F->Cluster); break;
        default            : /* should never arrive here */
                             return FAT_RET_CRITICAL_ERROR;
      }
      if (Res) return Res;
      #ifdef DEBUG
//      printf("Jump to cluster: %lu\n", F->Cluster);
      #endif
    }
  }
  return 0;
}


/* Tries to move the actual file position of the file up to the target */
/* position specified by the file pointer V->Files[Fd].TargetPos.      */
/* Returns 0 on success, or a negative error code on failure.          */
/* Called by fat_read (read.c) and fat_write (write.c) before every    */
/* byte transfer, and by truncate_or_extend (write.c) for extension.   */

//[KG] SPEEDUP: Try to move by MORE THAN ONE BYTE...
int move_to_targetpos(tFatVolume *V, int Fd, unsigned long BytesToWrite)
{
  DWORD     NewCluster;
  int       Res;
  tFatFile *F = &V->Files[Fd];
  unsigned long IncBy, MaxIncBy, NewFileSize;
  unsigned long OldFilePos, OldCluster, OldSectorInCluster, OldByteInSector, OldPrevCluster;

  NewFileSize = F->TargetPos+BytesToWrite;
  /* Check if we want to go before the beginning of the file */
  if (F->TargetPos < 0) return FD32_ERROR_INVALID_SEEK;

  /* If we are beyond the target position, we restart from the beginning */
  if (F->FilePos > F->TargetPos) {
    F->FilePos         = 0;
    F->Cluster         = F->FirstCluster;
    F->SectorInCluster = 0;
    F->ByteInSector    = 0;
  }

  /* If the file is zero bytes long it is a different story, that follows */
  if ((F->FirstCluster == 0) && (!(F->Mode & FAT_FMODE_ROOT)))
  {
    /* If we are reading from F, at EOF we just exit */
    if (!(F->Mode & FAT_FMODE_DO_WRITE)) return FAT_RET_EOF;
    /* If we are writing, we increment the file size and allocate */
    /* a new cluster that will be the file's first cluster.       */
/*    if (!(F->DirEntry.Attr & FD32_ATTR_DIRECTORY)) {
      IncBy = NewFileSize-F->DirEntry.FileSize;
      MaxIncBy = (V->Bpb.BPB_SecPerClus << 9);
      if (IncBy > MaxIncBy)
        IncBy = MaxIncBy;
      F->DirEntry.FileSize += IncBy;
    }*/
    Res = allocate_first_cluster(V, Fd, &NewCluster);
    if (Res) return Res;
    F->PrevCluster = F->Cluster = NewCluster;
  }
  if( (F->Mode & FAT_FMODE_DO_WRITE) &&
     !(F->DirEntry.Attr & FD32_ATTR_DIRECTORY) &&
      (NewFileSize > F->DirEntry.FileSize) ) { //we are writing to a file & it got to be increased
    // save file positioning information
    OldFilePos = F->FilePos;
    OldCluster = F->Cluster;
    OldSectorInCluster = F->SectorInCluster;
    OldByteInSector = F->ByteInSector;
    OldPrevCluster = F->PrevCluster;
    // move FilePos to end of File
    Res = advance_byte_position(V, Fd, 0); // This is just relevant if FilePos points to first byte of cluster
    if (Res) return Res;
    while (F->DirEntry.FileSize > F->FilePos) {
      IncBy = ((V->Bpb.BPB_SecPerClus-F->SectorInCluster) << 9)-F->ByteInSector; // jump to next cluster
      MaxIncBy = F->DirEntry.FileSize-F->FilePos;
      if (MaxIncBy < IncBy)
        IncBy = MaxIncBy;
      F->FilePos += IncBy;
      if (F->FilePos < F->DirEntry.FileSize) { // we need to jump to next cluster of cluster chain
        switch (V->FatType)
        {
          case FATTYPE_FAT12 : Res = fat12_read_entry(V, F->Cluster, 0, &F->Cluster); break;
          case FATTYPE_FAT16 : Res = fat16_read_entry(V, F->Cluster, 0, &F->Cluster); break;
          case FATTYPE_FAT32 : Res = fat32_read_entry(V, F->Cluster, 0, &F->Cluster); break;
          default            : /* should never arrive here */
                               return FAT_RET_CRITICAL_ERROR;
        }
        if (Res) return Res;
        F->ByteInSector = 0;
        F->SectorInCluster = 0;
      }
    }
    IncBy = (V->Bpb.BPB_SecPerClus << 9)-(((V->Bpb.BPB_SecPerClus << 9)-1) & F->DirEntry.FileSize);
    MaxIncBy = NewFileSize-F->DirEntry.FileSize;
    if (MaxIncBy < IncBy)
      IncBy = MaxIncBy;
    F->DirEntry.FileSize += IncBy;
    //increase FileSize
    while (F->DirEntry.FileSize < NewFileSize) {
      IncBy = (V->Bpb.BPB_SecPerClus << 9)-(((V->Bpb.BPB_SecPerClus << 9)-1) & F->DirEntry.FileSize);
      MaxIncBy = NewFileSize-F->DirEntry.FileSize;
      if (MaxIncBy < IncBy)
        IncBy = MaxIncBy;
      F->DirEntry.FileSize += IncBy;
//HIER NEUE CLUSTER ALLOZIERUNGSBEDINGUNG NOTWENDIG!
      //if (F->DirEntry.FileSize < NewFileSize) {
        F->PrevCluster = F->Cluster;
        Res = allocate_and_link_new_cluster(V, Fd, &NewCluster);
        if (Res) return Res; /* The disk may be full */
        F->Cluster = NewCluster;
      //}
    }
    //restore old file positioning information
    F->FilePos = OldFilePos;
    F->Cluster = OldCluster;
    F->SectorInCluster = OldSectorInCluster;
    F->ByteInSector = OldByteInSector;
    F->PrevCluster = OldPrevCluster;
  }

  /* And now perform the linear search from the current position */
  /* to the target position.                                     */
  int iLoopCount = 0;
  do {
    // [KG] calculate maximum no. of bytes to increment in this step
    IncBy = F->TargetPos-F->FilePos;
    MaxIncBy = V->Bpb.BPB_BytsPerSec - F->ByteInSector; //Increment by a maximum of ONE sector!
    if (MaxIncBy < IncBy)
      IncBy = MaxIncBy;

    if (F->FilePos != F->TargetPos) advance_byte_position(V, Fd, IncBy);
    if (end_of_file(V, Fd)) {
      /* If we are reading from the file, or the file is a FAT12/FAT16 */
      /* root directory, at EOF we just exit.                          */
      if ((!(F->Mode & FAT_FMODE_DO_WRITE)) || (F->Mode & FAT_FMODE_ROOT))
        return FAT_RET_EOF;

      /* But if we are writing, we increment the file size and allocate */
      /* a new cluster if we are at the last cluster boundary.          */
      if (F->DirEntry.Attr & FD32_ATTR_DIRECTORY) {
        if ((F->ByteInSector == 0) && (F->SectorInCluster == 0) && (BytesToWrite > 0) &&
            (F->FilePos != F->TargetPos)) {
          Res = allocate_and_link_new_cluster(V, Fd, &NewCluster);
          if (Res) return Res; /* The disk may be full */
          F->Cluster = NewCluster;
        }
      }
    }
    iLoopCount++;
    if ((iLoopCount % 20) == 0) RTX_Sleep_Time(1);
  }
  while (F->FilePos != F->TargetPos);
  return 0;
}


/*** FAT POSITIONING ANALYSIS ***

FilePos is the actual byte position inside the file;
FileSize is the file size in bytes, considered only if the file is not a dir;
TargetPos is the byte position to be reached, even out of the file extension;

Eof : for a file is FilePos==FileSize
      for a directory is Cluster==EOC (End Of Clusterchain)
      for a FAT12/FAT16 root directory is FilePos==BPB_RootEntCnt*32

1) FileSize>=0, 0<=FilePos<=FileSize, TargetPos<0;
Read, Write : exit reporting invalid condition (seek before the beginning)

2) FileSize>0, 0<=FilePos<=FileSize, FilePos==TargetPos;
Read  : exit if Eof, else do nothing
Write : if Eof, extend the file incrementing FileSize and allocating a new
        cluster if the boundary of the last cluster has been reached, unless
        the file is a FAT12/FAT16 root directory (which cannot be extended);
        if not Eof, do nothing

3) FileSize>0, 0<=FilePos<=FileSize, FilePos!=TargetPos;
Read  : advance FilePos until FilePos==TargetPos, exiting if Eof if reached
Write : advance FilePos until FilePos==TargetPos; while Eof if reached the
        file is extended, incrementing FileSize and allocating a new cluster
        if the boundary of the last cluster has been reached, unless the file
        is a FAT12/FAT16 root directory (which cannot be extended)

4) FileSize==0, FilePos==0, FilePos==TargetPos==0;
Read  : the same as case 2), but we are for sure at Eof
Write : since we are for sure at Eof, we need to extend the file, which we
        know is not a FAT12/FAT16 root directory; so we allocate a new cluster
        from scratch and set it as first cluster, and mark it into the FAT
        as last cluster of the chain; increment FileSize, next we continue
        with cases 2) and 3)

5) FileSize==0, FilePos==0, FilePos!=TargetPos>=0;
Read  : similar to case 3), but we have to exit immediatly because of Eof
Write : the same as case 4)

*/
