/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001-2002, Salvatore Isaja                               *
 *                                                                        *
 * This is "share.c" - Sharing conflict detection                         *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/
 
#include "fat1.h"

/* TO DO: This has not been tested yet!                      */
/*        I don't know if matrix index should be inverted... */

/* According to the RBIL:
 * (Table 1269) Values for DOS 7.x file sharing behavior
 * Flags for Second and subsequent Opens modes for each First Open mode,
 * encoded as 16-bit word as follows (using the notation of the RBIL):
 * N N N Y => 1
 * Y Y Y Y => F => 0x1F11
 * N N N Y => 1
 * N N N Y => 1
 */
static const WORD SharingBehaviour[5][5] =
{
  { 0xFEE9, 0x0000, 0x9009, 0x1111, 0x9999 },
  { 0x0000, 0x0000, 0x0000, 0x0000, 0x0000 },
  { 0x9009, 0x0000, 0x9009, 0x1911, 0x9999 },
  { 0x000F, 0x0000, 0x400F, 0x0401, 0x444F },
  { 0xF00F, 0x0000, 0xF00F, 0x1F11, 0xFFFF }
};


/* Given a file descriptor and the file descriptor of the first time the */
/* file has been opened, checks if reopening it is a valid operation.    */
/* Returns 0 on success, or a negative error code on failure.            */
/* FIX ME: According to the opening flag, should generate INT 24h on     */
/*         critical error.                                               */
/* Called by set_opening_mode (open.c).                                  */
int check_for_sharing_conflict(tFatVolume *V, int Fd, int FirstOpenFd)
{
  tFatFile *F  = &V->Files[Fd];
  tFatFile *F1 = &V->Files[FirstOpenFd];
  
  int FirstOpenMode, NextOpensMode, FirstOpenShare, NextOpensShare;

  switch (F1->Mode & FD32_OPEN_ACCESS_MASK)
  {
    case FD32_OPEN_ACCESS_READ      : FirstOpenMode = 0xF000; break;
    case FD32_OPEN_ACCESS_WRITE     : FirstOpenMode = 0x0F00; break;
    case FD32_OPEN_ACCESS_READWRITE : FirstOpenMode = 0x00F0; break;
    case FD32_OPEN_ACCESS_READ_NA   : FirstOpenMode = 0x000F; break;
    default                         : return FAT_RET_INVALID_SHARING;
  }

  switch (F->Mode & FD32_OPEN_ACCESS_MASK)
  {
    case FD32_OPEN_ACCESS_READ      : NextOpensMode = 0x8888; break;
    case FD32_OPEN_ACCESS_WRITE     : NextOpensMode = 0x4444; break;
    case FD32_OPEN_ACCESS_READWRITE : NextOpensMode = 0x2222; break;
    case FD32_OPEN_ACCESS_READ_NA   : NextOpensMode = 0x1111; break;
    default                         : return FAT_RET_INVALID_SHARING;
  }

  switch (F1->Mode & FD32_OPEN_SHARE_MASK)
  {
    case FD32_OPEN_SHARE_COMPAT    : FirstOpenShare = 0; break;
    case FD32_OPEN_SHARE_DENYALL   : FirstOpenShare = 1; break;
    case FD32_OPEN_SHARE_DENYWRITE : FirstOpenShare = 2; break;
    case FD32_OPEN_SHARE_DENYREAD  : FirstOpenShare = 3; break;
    case FD32_OPEN_SHARE_DENYNONE  : FirstOpenShare = 4; break;
    default                        : return FAT_RET_INVALID_SHARING;
  }

  switch (F->Mode & FD32_OPEN_SHARE_MASK)
  {
    case FD32_OPEN_SHARE_COMPAT    : NextOpensShare = 0; break;
    case FD32_OPEN_SHARE_DENYALL   : NextOpensShare = 1; break;
    case FD32_OPEN_SHARE_DENYWRITE : NextOpensShare = 2; break;
    case FD32_OPEN_SHARE_DENYREAD  : NextOpensShare = 3; break;
    case FD32_OPEN_SHARE_DENYNONE  : NextOpensShare = 4; break;
    default                        : return FAT_RET_INVALID_SHARING;
  }

  if ((SharingBehaviour[FirstOpenShare][NextOpensShare]
      & FirstOpenMode & NextOpensMode) == 0)
  {
    if (NextOpensShare == 0) return FAT_RET_SHARING_CRITICAL;
                             return FD32_ERROR_ACCESS_DENIED;
  }
  return 0;
}


/* Closes all instances of a file (identified by its first cluster) */
/* according to the DOS 5.0 sharing behaviour: if the file is open  */
/* in sharing compatibility mode it is closed, otherwise a sharing  */
/* violation critical error occurs.                                 */
/* Returns 0 on success, or a negative error code on failure.       */
/* Called by fat_unlink (unlink.c).                                 */
int close_shared_file(tFatVolume *V, DWORD FirstCluster)
{
  int k, Res;
  /* Check for open files not open in compatibility mode */
  for (k = 0; k < FAT_MAX_FILES; k++)
  {
    if (V->Files[k].Mode & FAT_FMODE_OPENED)
      if (V->Files[k].FirstCluster == FirstCluster)
        if ((V->Files[k].Mode & FD32_OPEN_SHARE_MASK) !=
            FD32_OPEN_SHARE_COMPAT) return FAT_RET_SHARING_CRITICAL;
  }
  /* The remaining files are open in compat mode and must be closed */
  for (k = 0; k < FAT_MAX_FILES; k++)
  {
    if (V->Files[k].Mode & FAT_FMODE_OPENED)
      if (V->Files[k].FirstCluster == FirstCluster)
      {
        Res = fat_close(V, k);
        if (Res) return Res;
      }
  }
  return 0;
}
