/**************************************************************************
 * FreeDOS 32 FAT Driver                                                  *
 * by Salvo Isaja                                                         *
 *                                                                        *
 * Copyright (C) 2001, Salvatore Isaja                                    *
 *                                                                        *
 * This is "shortnam.c" - Support code for short file names (8.3 format)  *
 *                                                                        *
 *                                                                        *
 * This file is part of the FreeDOS 32 FAT Driver.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is free software; you can redistribute it    *
 * and/or modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2 of the  *
 * License, or (at your option) any later version.                        *
 *                                                                        *
 * The FreeDOS 32 FAT Driver is distributed in the hope that it will be   *
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 * GNU General Public License for more details.                           *
 *                                                                        *
 * You should have received a copy of the GNU General Public License      *
 * along with the FreeDOS 32 FAT Driver; see the file COPYING;            *
 * if not, write to the Free Software Foundation, Inc.,                   *
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA                *
 **************************************************************************/

#include "fat1.h"

/* Checks if a character Ch is valid for a FAT short file name.      */
/* Returns the valid character if successful, or 0 if Ch is invalid. */
/* Called by validate_short_file_name.                               */
static char validate_character(char Ch)
{
  /* Lower case letters are not valid */
  if ((Ch >= 'a') && (Ch <= 'z')) return Ch - 'a' + 'A';
#if 0
  /* Convert extended ASCII codes */
  if (Ch >= 128) return ShortNameEqu[Ch-128];
#endif
  /* The following characters are not valid */
  if (Ch < 0x20) return 0;
  switch (Ch)
  {
    case 0x22 : /* " */
    case 0x2A : /* * */
    case 0x2B : /* + */
    case 0x2C : /* , */
    case 0x2E : /* . */
    case 0x2F : /* / */
    case 0x3A : /* : */
    case 0x3B : /* ; */
    case 0x3C : /* < */
    case 0x3D : /* = */
    case 0x3E : /* > */
    case 0x3F : /* ? */
    case 0x5B : /* [ */
    case 0x5C : /* \ */
    case 0x5D : /* ] */
    case 0x7C : /* | */ return 0;
  }
  return Ch;
}


/* Given a file name Source, checks if it's valid and return in      */
/* Dest[11] the absolute name in the format of the directory entry:  */
/* - 8 characters for name, trailed with padding spaces              */
/* - 3 characters for extension, trailed with padding spaces         */
/* - no separation dot or invalid characters                         */
/* Note: directory names "dot" (.) and "dotdot" (..) are exceptions. */
/* Returns 0 on success, or a negative error code on failure.        */
/* Called by allocate_sfn_dir_entry (direntry.c) and                 */
/* allocate_lfn_dir_entries (direntry.c).                            */
int validate_short_file_name(char *Dest, char *Source)
{
  int k;

  /* Name and extension have to be padded with spaces */
  memset(Dest,0x20,11);
  
  /* Build ".          " and "..         " if Source is "." or ".." */
  if ((strcmp(Source,".") == 0) || (strcmp(Source,"..") == 0))
  {
    for (;*Source;Source++,Dest++) *Dest = *Source;
    return 0;
  }

  /* Copy the name, un upper case, into the first 8 characters of Dest */
  for (k=0; *Source != '.'; k++, Source++)
  {
    if (*Source == 0) return 0;
    if (k > 7) break;
    Dest[k] = validate_character(*Source);
    if (Dest[k] == 0) return FAT_RET_INVALID_FILENAME;
  }
  if ((unsigned char) Dest[0] == 0xE5)
    Dest[0] = FAT_DIR_E5; /* 0xE5 would be FAT_DIR_FREE */

  /* Search for the extension */
  while (*Source != '.')
  {
    if (*Source == 0) return 0;
    Source++;
  }
  Source++;

  /* Copy the extension, in upper case, into the last 3 characters of Dest */
  for (k=8; *Source != 0; k++, Source++)
  {
    if (k > 10) break;
    Dest[k] = validate_character(*Source);
    if (Dest[k] == 0) return FAT_RET_INVALID_FILENAME;
  }

  /* The final check: DIR_Name[0] cannot be 0x20 (space) */
  if (Dest[0] == 0x20) return FAT_RET_INVALID_FILENAME;
  return 0;
}


/* Gets an ASCIZ short file name from an 11-bytes absolute name in   */
/* the format of the directory entry:                                */
/* - 8 characters for name, trailed with padding spaces              */
/* - 3 characters for extension, trailed with padding spaces         */
/* - no separation dot or invalid characters                         */
/* Called by find (find.c).                                          */
void absolute_name_to_short_name(char *Dest, char *Source)
{
  char *NameEnd;
  char *ExtEnd;
  char *s;

  /* Remove padding spaces from the name and the extension */
  for (NameEnd = Source+7;  *NameEnd == ' '; NameEnd--);
  for (ExtEnd  = Source+10; *ExtEnd  == ' '; ExtEnd--);
  
  /* Put the name, a dot and the extension in Dest */
  s = Source;
  if (*s == FAT_DIR_E5) *(Dest++) = 0xE5, s++;
  for (; s <= NameEnd; Dest++, s++) *Dest = *s;
  if (Source+8 <= ExtEnd) *(Dest++) = '.';
  for (s = Source+8; s <= ExtEnd; Dest++, s++) *Dest = *s;
  *Dest = 0;
  if (Dest[0] == FAT_DIR_E5) Dest[0] = 0xE5; /* 0xE5 would be FAT_DIR_FREE */
}

// [KG] Gets an 11-byte absolute name from an ASCIZ short filename
void short_name_to_absolute_name(char *source, char *dest)
{
  int  i=0;
  char c;
  bool isSpecDir;
  isSpecDir = (source[0] == '.');
  memset(dest, ' ', 11);
  if (strlen(source)) {
    while (((c=*source)!= 0) && (i <= 11)) {
      if ((c == '.') && (!isSpecDir))
        i = 8;
      else
        dest[i++] = toupper(c);
      source++;
    }
  }
}

